﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;

namespace Skeleton_Demo
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();
            ofd = new OpenFileDialog();
            sfd = new SaveFileDialog();
            string filter = "Image|*.jpg;*.png;*.gif";
            ofd.Filter = filter;
            sfd.Filter = filter;
            steps = new List<byte[]>();
        }

        OpenFileDialog ofd;
        SaveFileDialog sfd;
        string image_path;
        Bitmap image, processed, display;
        public static List<byte[]> steps;

        private void skeletonizeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            processed = processed.Skeletonize();
            display = processed.ScaleImage(image_frame.Width, image_frame.Height);
            image_frame.Image = display;
        }

        private void saveToolStripMenuItem_Click(object sender, EventArgs e)
        {
            sfd.FileName = image_path;
            if (sfd.ShowDialog() == DialogResult.OK)
            {
                processed.Save(sfd.FileName);
            }
        }

        private void loadToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (ofd.ShowDialog() == DialogResult.OK)
            {
                image_path = ofd.FileName;
                image = (Bitmap)Bitmap.FromFile(image_path);
                processed = image;
                display = processed.ScaleImage(image_frame.Width, image_frame.Height);
                image_frame.Image = display;
            }
        }
    }

    public static class Ext
    {
        public static Bitmap ScaleImage(this Bitmap image, int width, int height)
        {
            float xratio = (float)width / image.Width;
            float yratio = (float)height / image.Height;
            float ratio = Math.Min(xratio, yratio);

            int w = (int)(image.Width * ratio);
            int h = (int)(image.Height * ratio);

            Bitmap result = new Bitmap(image, new Size(w, h));
            return result;
        }

        public static byte[] Erode(this byte[] buffer, int width, int height)
        {
            byte[] result = new byte[buffer.Length];

            for (int x = 1; x < width - 1; x++)
            {
                for (int y = 1; y < height - 1; y++)
                {
                    int position = x + y * width;
                    byte val = 15;
                    for (int i = -1; i < 2; i++)
                    {
                        for (int j = -1; j < 2; j++)
                        {
                            int se_pos = position + i + j * width;
                            val = Math.Min(val, buffer[se_pos]);
                        }
                    }
                    result[position] = val;
                }
            }

            return result;
        }

        public static byte[] Dilate(this byte[] buffer, int width, int height)
        {
            byte[] result = new byte[buffer.Length];

            for (int x = 1; x < width - 1; x++)
            {
                for (int y = 1; y < height - 1; y++)
                {
                    int position = x + y * width;
                    byte val = 0;
                    for (int i = -1; i < 2; i++)
                    {
                        for (int j = -1; j < 2; j++)
                        {
                            int se_pos = position + i + j * width;
                            val = Math.Max(val, buffer[se_pos]);
                        }
                    }
                    result[position] = val;
                }
            }

            return result;
        }

        public static Bitmap Skeletonize(this Bitmap image)
        {
            int w = image.Width;
            int h = image.Height;

            BitmapData image_data = image.LockBits(
                new Rectangle(0, 0, w, h),
                ImageLockMode.ReadOnly,
                PixelFormat.Format8bppIndexed);

            int bytes = image_data.Stride * image_data.Height;
            byte[] buffer = new byte[bytes];
            byte[] temp = new byte[bytes];
            byte[] result = new byte[bytes];

            Marshal.Copy(image_data.Scan0, buffer, 0, bytes);
            image.UnlockBits(image_data);

            while (true)
            {
                temp = buffer.Erode(w, h);
                int sum = temp.Sum(x => (int)x);
                if (sum == 0)
                {
                    break;
                }
                temp = temp.Dilate(w, h);
                for (int i = 0; i < bytes; i++)
                {
                    result[i] += (byte)(buffer[i] - temp[i]);
                }
                buffer = buffer.Erode(w, h);
            }

            Bitmap res_img = new Bitmap(w, h);
            BitmapData res_data = res_img.LockBits(
                new Rectangle(0, 0, w, h),
                ImageLockMode.WriteOnly,
                PixelFormat.Format8bppIndexed);
            Marshal.Copy(result, 0, res_data.Scan0, bytes);
            res_img.UnlockBits(res_data);

            return res_img;
        }
    }
}
